# -*- encoding: utf-8 -*-
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief Code generated by Mobius. Do not edit
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
EXTENSION_ID = 'datasource-scsi'
EXTENSION_NAME = 'Datasource SCSI'
EXTENSION_AUTHOR = 'Vladimir Santos'
EXTENSION_VERSION = '0.1.1'
EXTENSION_DESCRIPTION = 'SCSI datasource support'

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# Mobius Forensic Toolkit
# Copyright (C) 2008,2009,2010 Eduardo Aguiar
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
import os.path
import os
import urllib
import fcntl
import struct
import ctypes

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief constants
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
STATUS_GOOD, STATUS_BAD, STATUS_UNKNOWN = range (3)
SG_DXFER_FROM_DEV = -3                                          # <scsi/sg.h>
SG_IO = 0x2285                                                  # <scsi/sg.h>
SG_INFO_OK_MASK = 0x01                                          # <scsi/sg.h>
SG_INFO_OK = 0x00                                               # <scsi/sg.h>
SCSI_INQUIRY = 0x12                                             # <scsi/scsi.h>
SCSI_READ_CAPACITY = 0x25                                       # <scsi/scsi.h>
VPD_SUPPORTED_VPDS = 0x00
VPD_UNIT_SERIAL_NUM = 0x80

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief SCSI Reader
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class Reader (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize reader
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__ (self, datasource):
    self.__uri = gdata.mediator.call ('object.get-attribute', datasource, 'uri')
    self.__image_size = int (gdata.mediator.call ('object.get-attribute', datasource, 'size'))
    self.__sector_size = int (gdata.mediator.call ('object.get-attribute', datasource, 'sector_size'))
    self.__pos = 0
    self.__stream = None

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief del reader
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __del__ (self):
    self.close ()

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @deprecated walk through all segment files
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __iter__ (self):
    data = self.read (self.__sector_size)
    while data:
      yield STATUS_UNKNOWN, data
      data = self.read (self.__sector_size)

    self.close ()

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief close reader
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def close (self):
    if self.__stream:
      self.__stream.close ()
      self.__image_size = 0

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief set reading position
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def seek (self, pos, whence):
    if whence == 0:	# from beginning of file
      abs_pos = pos

    elif whence == 1:	# from current position
      abs_pos = self.__pos + pos

    elif whence == 2:	# from end of file
      abs_pos = self.__image_size - 1 + pos

    if abs_pos >= 0 and abs_pos < self.__image_size:
      self.__pos = abs_pos
      self.__stream = None

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief get the current reading position
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def tell (self):
    return self.__pos

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief read bytes
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def read (self, siz):
    if not self.__stream:
      path = urllib.unquote (self.__uri)[7:]
      self.__stream = open (path)
      self.__stream.seek (self.__pos, 0)

    data = self.__stream.read (siz)
    self.__pos += len (data)

    return data

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief return stream size
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __get_size (self):
    return self.__image_size

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # properties
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  size = property (__get_size, None)


# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief SCSI datasource support
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class SCSIHandler (object):

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief initialize object
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def __init__ (self):
    object.__init__ (self)
    self.type = 'scsi'
    self.description = 'SCSI device'
    self.extensions = []
    self.Reader = Reader
    self.default_item_category = 'harddisk'
    self.is_exportable = False
    
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief verify if it is a linux device
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def signature_found (self, uri):
    path = urllib.unquote(uri)[7:]
    statinfo = os.stat(path)

    if (((statinfo.st_mode & 0x06000) == 0x6000) and    # it is a block device
        (os.major(statinfo.st_rdev) == 8) and           # SCSI disk device type
        ((os.minor(statinfo.st_rdev) & 0x0f) == 0)):    # it not a partition
      return True

    return False

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief retrieve metadata from uri
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def retrieve_metadata (self, uri):
    path = urllib.unquote(uri)[7:]

    metadata = {}
    metadata['type'] = 'scsi'
    metadata['uri'] = uri
    metadata['name'] = os.path.basename(path)    # device name

    fd = os.open(path, os.O_RDONLY)

    # SCSI inquiry
    cdb = ctypes.create_string_buffer(6)
    sense = ctypes.create_string_buffer(32)
    buf = ctypes.create_string_buffer(128)
    io = struct.pack('@iiBB2xIPPPI8xP20x', ord('S'), SG_DXFER_FROM_DEV,
                     len(cdb), len(sense), len(buf), ctypes.addressof(buf),
                     ctypes.addressof(cdb), ctypes.addressof(sense), 10000, 0)
    sg_io = ctypes.create_string_buffer(len(io))
    sg_io.value = io
    cdb[0] = chr(SCSI_INQUIRY)
    if fcntl.ioctl(fd, SG_IO, sg_io) >= 0:
      if (struct.unpack('@I', sg_io[-4:])[0] & SG_INFO_OK_MASK) == SG_INFO_OK:
        metadata['drive_vendor'] = buf[8:16].strip()
        metadata['drive_model'] = buf[16:32].strip()
        metadata['drive_firmware'] = buf[32:36].strip()

    cdb[1] = chr(0x01)                           # read vital product data
    cdb[2] = chr(VPD_SUPPORTED_VPDS)             # supported vpd's page code
    if fcntl.ioctl(fd, SG_IO, sg_io) >= 0:
      if (struct.unpack('@I', sg_io[-4:])[0] & SG_INFO_OK_MASK) == SG_INFO_OK:
        if chr(VPD_UNIT_SERIAL_NUM) in buf[4:(4 + ord(buf[3]))]:
          cdb[2] = chr(VPD_UNIT_SERIAL_NUM)      # unit serial number page code
          if fcntl.ioctl(fd, SG_IO, sg_io) >= 0:
            if (struct.unpack('@I', sg_io[-4:])[0] & SG_INFO_OK_MASK) == SG_INFO_OK:
              metadata['drive_serial'] = buf[4:(4 + ord(buf[3]))].strip()

    # SCSI read capacity (10)
    cdb = ctypes.create_string_buffer(10)
    sense = ctypes.create_string_buffer(32)
    buf = ctypes.create_string_buffer(128)
    io = struct.pack('@iiBB2xIPPPI8xP20x', ord('S'), SG_DXFER_FROM_DEV,
                     len(cdb), len(sense), len(buf), ctypes.addressof(buf),
                     ctypes.addressof(cdb), ctypes.addressof(sense), 10000, 0)
    sg_io = ctypes.create_string_buffer(len(io))
    sg_io.value = io
    cdb[0] = chr(SCSI_READ_CAPACITY)
    if fcntl.ioctl(fd, SG_IO, sg_io) >= 0:
      if (struct.unpack('@I', sg_io[-4:])[0] & SG_INFO_OK_MASK) == SG_INFO_OK:
        last_lba, block_len = struct.unpack('>II', buf[:8])
        if last_lba < 0xffffffff:
          last_lba = last_lba + 1
        metadata['total_sectors'] = last_lba
        metadata['sector_size'] = block_len
        metadata['size'] = last_lba * block_len

    os.close(fd)

    return metadata

  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  # @brief set item attributes
  # =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  def set_item_attributes (self, item, datasource):
    category = gdata.mediator.call ('item.get-attribute', item, 'category')
    if category != 'harddisk':
      return

    drive_vendor = gdata.mediator.call ('object.get-attribute', datasource, 'drive_vendor')
    part_id = gdata.mediator.call ('object.get-attribute', datasource, 'drive_model')
    drive_serial = gdata.mediator.call ('object.get-attribute', datasource, 'drive_serial')
    drive_firmware = gdata.mediator.call ('object.get-attribute', datasource, 'drive_firmware')
    size = gdata.mediator.call ('object.get-attribute', datasource, 'size')
    total_sectors = gdata.mediator.call ('object.get-attribute', datasource, 'total_sectors')
    sector_size = gdata.mediator.call ('object.get-attribute', datasource, 'sector_size')

    gdata.mediator.call ('item.set-attribute', item, 'vendor', drive_vendor)
    gdata.mediator.call ('item.set-attribute', item, 'part_id', part_id)
    gdata.mediator.call ('item.set-attribute', item, 'serial', drive_serial)
    gdata.mediator.call ('item.set-attribute', item, 'firmware', drive_firmware)
    gdata.mediator.call ('item.set-attribute', item, 'lba', total_sectors)
    gdata.mediator.call ('item.set-attribute', item, 'sector_size', sector_size)
    gdata.mediator.call ('item.set-attribute', item, 'real_capacity', size)
    gdata.mediator.call ('item.expand-value-masks', item)

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief start function
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_start ():
  gdata.mediator.call ('datasource.add-handler', SCSIHandler ())

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# @brief stop function
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_stop ():
  gdata.mediator.call ('datasource.remove-handler', 'scsi')

# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
# API initialization
# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
def pvt_start_api ():
  pass
